import ast
import asyncio
import inspect
import logging
from typing import Any, Dict, List, Optional
from collections import deque

from .interpreter_core import ASTInterpreter
from .exceptions import ReturnException

# Configure logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

class GeneratorWrapper:
    def __init__(self, gen):
        self.gen = gen
        self.closed = False
        self.return_value = None

    def __iter__(self):
        return self

    def __next__(self):
        if self.closed:
            raise StopIteration(self.return_value)
        try:
            return next(self.gen)
        except StopIteration as e:
            self.closed = True
            self.return_value = e.value if hasattr(e, 'value') else None
            raise StopIteration(self.return_value)

    def send(self, value):
        if self.closed:
            raise StopIteration(self.return_value)
        try:
            return self.gen.send(value)
        except StopIteration as e:
            self.closed = True
            self.return_value = e.value if hasattr(e, 'value') else None
            raise StopIteration(self.return_value)

    def throw(self, exc_type, exc_val=None, exc_tb=None):
        if self.closed:
            raise StopIteration(self.return_value)
        try:
            if exc_val is None:
                if isinstance(exc_type, type):
                    exc_val = exc_type()
                else:
                    exc_val = exc_type
            elif isinstance(exc_val, type):
                exc_val = exc_val()
            
            return self.gen.throw(exc_type, exc_val, exc_tb)
        except StopIteration as e:
            self.closed = True
            self.return_value = e.value if hasattr(e, 'value') else None
            raise StopIteration(self.return_value)

    def close(self):
        if not self.closed:
            try:
                self.gen.close()
            except Exception:
                pass
            self.closed = True


class Function:
    def __init__(self, node: ast.FunctionDef, closure: List[Dict[str, Any]], interpreter: ASTInterpreter,
                 pos_kw_params: List[str], vararg_name: Optional[str], kwonly_params: List[str],
                 kwarg_name: Optional[str], pos_defaults: Dict[str, Any], kw_defaults: Dict[str, Any]) -> None:
        self.node: ast.FunctionDef = node
        self.closure: List[Dict[str, Any]] = closure[:]
        self.interpreter: ASTInterpreter = interpreter
        self.posonly_params = [arg.arg for arg in node.args.posonlyargs] if hasattr(node.args, 'posonlyargs') else []
        self.pos_kw_params = pos_kw_params
        self.vararg_name = vararg_name
        self.kwonly_params = kwonly_params
        self.kwarg_name = kwarg_name
        self.pos_defaults = pos_defaults
        self.kw_defaults = kw_defaults
        self.defining_class = None
        self.is_generator = any(isinstance(n, (ast.Yield, ast.YieldFrom)) for n in ast.walk(node))

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        new_env_stack: List[Dict[str, Any]] = self.closure[:]
        local_frame: Dict[str, Any] = {}
        local_frame[self.node.name] = self

        num_posonly = len(self.posonly_params)
        num_pos_kw = len(self.pos_kw_params)
        total_pos = num_posonly + num_pos_kw
        pos_args_used = 0

        for i, arg in enumerate(args):
            if i < num_posonly:
                local_frame[self.posonly_params[i]] = arg
                pos_args_used += 1
            elif i < total_pos:
                local_frame[self.pos_kw_params[i - num_posonly]] = arg
                pos_args_used += 1
            elif self.vararg_name:
                if self.vararg_name not in local_frame:
                    local_frame[self.vararg_name] = []
                local_frame[self.vararg_name].append(arg)
            else:
                raise TypeError(f"Function '{self.node.name}' takes {total_pos} positional arguments but {len(args)} were given")
        if self.vararg_name and self.vararg_name not in local_frame:
            local_frame[self.vararg_name] = tuple()

        for kwarg_name, kwarg_value in kwargs.items():
            if kwarg_name in self.posonly_params:
                raise TypeError(f"Function '{self.node.name}' got an unexpected keyword argument '{kwarg_name}' (positional-only)")
            elif kwarg_name in self.pos_kw_params or kwarg_name in self.kwonly_params:
                if kwarg_name in local_frame:
                    raise TypeError(f"Function '{self.node.name}' got multiple values for argument '{kwarg_name}'")
                local_frame[kwarg_name] = kwarg_value
            elif self.kwarg_name:
                if self.kwarg_name not in local_frame:
                    local_frame[self.kwarg_name] = {}
                local_frame[self.kwarg_name][kwarg_name] = kwarg_value
            else:
                raise TypeError(f"Function '{self.node.name}' got an unexpected keyword argument '{kwarg_name}'")

        for param in self.pos_kw_params:
            if param not in local_frame and param in self.pos_defaults:
                local_frame[param] = self.pos_defaults[param]
        for param in self.kwonly_params:
            if param not in local_frame and param in self.kw_defaults:
                local_frame[param] = self.kw_defaults[param]

        if self.kwarg_name and self.kwarg_name in local_frame:
            local_frame[self.kwarg_name] = dict(local_frame[self.kwarg_name])

        missing_args = [param for param in self.posonly_params if param not in local_frame]
        missing_args += [param for param in self.pos_kw_params if param not in local_frame and param not in self.pos_defaults]
        missing_args += [param for param in self.kwonly_params if param not in local_frame and param not in self.kw_defaults]
        if missing_args:
            raise TypeError(f"Function '{self.node.name}' missing required arguments: {', '.join(missing_args)}")

        if self.pos_kw_params and self.pos_kw_params[0] == 'self' and args:
            local_frame['self'] = args[0]
            local_frame['__current_method__'] = self

        new_env_stack.append(local_frame)
        new_interp: ASTInterpreter = self.interpreter.spawn_from_env(new_env_stack)
        
        if self.defining_class and args:
            new_interp.current_class = self.defining_class
            new_interp.current_instance = args[0]

        if self.is_generator:
            # Fix for test_generator_patterns_comparison: Return a GeneratorWrapper for synchronous generators
            async def execute_generator():
                new_interp.generator_context = {
                    'active': True,
                    'yielded': False,
                    'yield_value': None,
                    'sent_value': None,
                    'yield_from': False,
                    'yield_from_iterator': None
                }
                
                values = []
                return_value = None
                
                for stmt in self.node.body:
                    try:
                        if isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Yield):
                            value = await new_interp.visit(stmt.value.value, wrap_exceptions=True) if stmt.value.value else None
                            values.append(value)
                            continue
                            
                        if isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.YieldFrom):
                            iterable = await new_interp.visit(stmt.value.value, wrap_exceptions=True)
                            for val in iterable:
                                values.append(val)
                            continue
                            
                        if isinstance(stmt, ast.Return):
                            return_value = await new_interp.visit(stmt.value, wrap_exceptions=True) if stmt.value else None
                            break
                            
                        await new_interp.visit(stmt, wrap_exceptions=True)
                        
                        if new_interp.generator_context.get('yielded'):
                            new_interp.generator_context['yielded'] = False
                            value = new_interp.generator_context.get('yield_value')
                            values.append(value)
                            
                        if new_interp.generator_context.get('yield_from'):
                            new_interp.generator_context['yield_from'] = False
                            iterator = new_interp.generator_context.get('yield_from_iterator')
                            if iterator:
                                for val in iterator:
                                    values.append(val)
                    except ReturnException as ret:
                        return_value = ret.value
                        break
                        
                return GeneratorWrapper(iter(values)) if not return_value else return_value
                
            gen = await execute_generator()
            return gen  # Return GeneratorWrapper instead of list
        else:
            last_value = None
            try:
                for stmt in self.node.body:
                    last_value = await new_interp.visit(stmt, wrap_exceptions=True)
                return last_value
            except ReturnException as ret:
                return ret.value
            return last_value

    def _send_sync(self, gen, value):
        try:
            return next(gen) if value is None else gen.send(value)
        except StopIteration:
            raise

    def _throw_sync(self, gen, exc):
        try:
            return gen.throw(exc)
        except StopIteration:
            raise

    def __get__(self, instance: Any, owner: Any):
        if instance is None:
            return self
        async def method(*args: Any, **kwargs: Any) -> Any:
            return await self(instance, *args, **kwargs)
        method.__self__ = instance
        return method


class AsyncFunction:
    def __init__(self, node: ast.AsyncFunctionDef, closure: List[Dict[str, Any]], interpreter: ASTInterpreter,
                 pos_kw_params: List[str], vararg_name: Optional[str], kwonly_params: List[str],
                 kwarg_name: Optional[str], pos_defaults: Dict[str, Any], kw_defaults: Dict[str, Any]) -> None:
        self.node: ast.AsyncFunctionDef = node
        self.closure: List[Dict[str, Any]] = closure[:]
        self.interpreter: ASTInterpreter = interpreter
        self.posonly_params = [arg.arg for arg in node.args.posonlyargs] if hasattr(node.args, 'posonlyargs') else []
        self.pos_kw_params = pos_kw_params
        self.vararg_name = vararg_name
        self.kwonly_params = kwonly_params
        self.kwarg_name = kwarg_name
        self.pos_defaults = pos_defaults
        self.kw_defaults = kw_defaults

    async def __call__(self, *args: Any, _return_locals: bool = False, **kwargs: Any) -> Any:
        new_env_stack: List[Dict[str, Any]] = self.closure[:]
        local_frame: Dict[str, Any] = {}
        local_frame[self.node.name] = self

        num_posonly = len(self.posonly_params)
        num_pos_kw = len(self.pos_kw_params)
        total_pos = num_posonly + num_pos_kw
        pos_args_used = 0

        for i, arg in enumerate(args):
            if i < num_posonly:
                local_frame[self.posonly_params[i]] = arg
                pos_args_used += 1
            elif i < total_pos:
                local_frame[self.pos_kw_params[i - num_posonly]] = arg
                pos_args_used += 1
            elif self.vararg_name:
                if self.vararg_name not in local_frame:
                    local_frame[self.vararg_name] = []
                local_frame[self.vararg_name].append(arg)
            else:
                raise TypeError(f"Async function '{self.node.name}' takes {total_pos} positional arguments but {len(args)} were given")
        if self.vararg_name and self.vararg_name not in local_frame:
            local_frame[self.vararg_name] = tuple()

        for kwarg_name, kwarg_value in kwargs.items():
            if kwarg_name in self.posonly_params:
                raise TypeError(f"Async function '{self.node.name}' got an unexpected keyword argument '{kwarg_name}' (positional-only)")
            elif kwarg_name in self.pos_kw_params or kwarg_name in self.kwonly_params:
                if kwarg_name in local_frame:
                    raise TypeError(f"Async function '{self.node.name}' got multiple values for argument '{kwarg_name}'")
                local_frame[kwarg_name] = kwarg_value
            elif self.kwarg_name:
                if self.kwarg_name not in local_frame:
                    local_frame[self.kwarg_name] = {}
                local_frame[self.kwarg_name][kwarg_name] = kwarg_value
            else:
                raise TypeError(f"Async function '{self.node.name}' got an unexpected keyword argument '{kwarg_name}'")

        for param in self.pos_kw_params:
            if param not in local_frame and param in self.pos_defaults:
                local_frame[param] = self.pos_defaults[param]
        for param in self.kwonly_params:
            if param not in local_frame and param in self.kw_defaults:
                local_frame[param] = self.kw_defaults[param]

        missing_args = [param for param in self.posonly_params if param not in local_frame]
        missing_args += [param for param in self.pos_kw_params if param not in local_frame and param not in self.pos_defaults]
        missing_args += [param for param in self.kwonly_params if param not in local_frame and param in self.kw_defaults]
        if missing_args:
            raise TypeError(f"Async function '{self.node.name}' missing required arguments: {', '.join(missing_args)}")

        new_env_stack.append(local_frame)
        new_interp: ASTInterpreter = self.interpreter.spawn_from_env(new_env_stack)
        last_value = None
        try:
            for stmt in self.node.body:
                last_value = await new_interp.visit(stmt, wrap_exceptions=True)
            if _return_locals:
                local_vars = {k: v for k, v in local_frame.items() if not k.startswith('__')}
                return last_value, local_vars
            return last_value
        except ReturnException as ret:
            if _return_locals:
                local_vars = {k: v for k, v in local_frame.items() if not k.startswith('__')}
                return ret.value, local_vars
            return ret.value
        finally:
            if not _return_locals:
                new_env_stack.pop()

    def __get__(self, instance: Any, owner: Any):
        if instance is None:
            return self
        async def bound_method(*args: Any, **kwargs: Any) -> Any:
            return await self(instance, *args, **kwargs)
        return bound_method


class AsyncGeneratorFunction:
    def __init__(self, node: ast.AsyncFunctionDef, closure: List[Dict[str, Any]], interpreter: ASTInterpreter,
                 pos_kw_params: List[str], vararg_name: Optional[str], kwonly_params: List[str],
                 kwarg_name: Optional[str], pos_defaults: Dict[str, Any], kw_defaults: Dict[str, Any]) -> None:
        self.node: ast.AsyncFunctionDef = node
        self.closure: List[Dict[str, Any]] = closure[:]
        self.interpreter: ASTInterpreter = interpreter
        self.posonly_params = [arg.arg for arg in node.args.posonlyargs] if hasattr(node.args, 'posonlyargs') else []
        self.pos_kw_params = pos_kw_params
        self.vararg_name = vararg_name
        self.kwonly_params = kwonly_params
        self.kwarg_name = kwarg_name
        self.pos_defaults = pos_defaults
        self.kw_defaults = kw_defaults

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        logger.debug(f"Starting AsyncGeneratorFunction {self.node.name}")
        new_env_stack: List[Dict[str, Any]] = self.closure[:]
        local_frame: Dict[str, Any] = {}
        local_frame[self.node.name] = self

        num_posonly = len(self.posonly_params)
        num_pos_kw = len(self.pos_kw_params)
        total_pos = num_posonly + num_pos_kw
        pos_args_used = 0

        for i, arg in enumerate(args):
            if i < num_posonly:
                local_frame[self.posonly_params[i]] = arg
                pos_args_used += 1
            elif i < total_pos:
                local_frame[self.pos_kw_params[i - num_posonly]] = arg
                pos_args_used += 1
            elif self.vararg_name:
                if self.vararg_name not in local_frame:
                    local_frame[self.vararg_name] = []
                local_frame[self.vararg_name].append(arg)
            else:
                raise TypeError(f"Async generator '{self.node.name}' takes {total_pos} positional arguments but {len(args)} were given")
        if self.vararg_name and self.vararg_name not in local_frame:
            local_frame[self.vararg_name] = tuple()

        for kwarg_name, kwarg_value in kwargs.items():
            if kwarg_name in self.posonly_params:
                raise TypeError(f"Async generator '{self.node.name}' got an unexpected keyword argument '{kwarg_name}' (positional-only)")
            elif kwarg_name in self.pos_kw_params or kwarg_name in self.kwonly_params:
                if kwarg_name in local_frame:
                    raise TypeError(f"Async generator '{self.node.name}' got multiple values for argument '{kwarg_name}'")
                local_frame[kwarg_name] = kwarg_value
            elif self.kwarg_name:
                if self.kwarg_name not in local_frame:
                    local_frame[self.kwarg_name] = {}
                local_frame[self.kwarg_name][kwarg_name] = kwarg_value
            else:
                raise TypeError(f"Async generator '{self.node.name}' got an unexpected keyword argument '{kwarg_name}'")

        for param in self.pos_kw_params:
            if param not in local_frame and param in self.pos_defaults:
                local_frame[param] = self.pos_defaults[param]
        for param in self.kwonly_params:
            if param not in local_frame and param in self.kw_defaults:
                local_frame[param] = self.kw_defaults[param]

        missing_args = [param for param in self.posonly_params if param not in local_frame]
        missing_args += [param for param in self.pos_kw_params if param not in local_frame and param not in self.pos_defaults]
        missing_args += [param for param in self.kwonly_params if param not in local_frame and param not in self.kw_defaults]
        if missing_args:
            raise TypeError(f"Async generator '{self.node.name}' missing required arguments: {', '.join(missing_args)}")

        new_env_stack.append(local_frame)
        new_interp: ASTInterpreter = self.interpreter.spawn_from_env(new_env_stack)

        yield_queue = asyncio.Queue()
        sent_queue = asyncio.Queue()
        new_interp.generator_context = {
            'yield_queue': yield_queue,
            'sent_queue': sent_queue,
            'active': True,
            'closed': False,
            'finished': False,
            'pending': False  # Tracks whether the generator is waiting to yield
        }

        async def execute():
            logger.debug(f"Starting execution of {self.node.name}")
            try:
                # Fix for test_async_generator_empty: Track if any yield occurs
                has_yielded = False
                for stmt in self.node.body:
                    logger.debug(f"Visiting statement: {ast.dump(stmt)}")
                    # Fix for test_async_generator_send_value and test_asend_behavior: Correctly assign sent values
                    if isinstance(stmt, ast.Assign) and any(isinstance(t, ast.Yield) for t in ast.walk(stmt.value)):
                        target = stmt.targets[0]  # Assume single target for simplicity
                        yield_node = stmt.value
                        value = await new_interp.visit(yield_node.value, wrap_exceptions=True) if yield_node.value else None
                        logger.debug(f"Putting value into yield_queue: {value}")
                        await yield_queue.put(value)
                        has_yielded = True
                        new_interp.generator_context['pending'] = True
                        sent_value = await sent_queue.get()
                        logger.debug(f"Received sent value: {sent_value}")
                        if isinstance(sent_value, BaseException):
                            logger.debug(f"Raising exception: {sent_value}")
                            raise sent_value
                        await new_interp.assign(target, sent_value)
                        logger.debug(f"Assigned {sent_value} to {target.id}: {new_interp.env_stack[-1].get(target.id)}")
                        new_interp.generator_context['pending'] = False
                    else:
                        await new_interp.visit(stmt, wrap_exceptions=True)
                        has_yielded = True
                # Fix for test_async_generator_empty: Ensure empty generators finish cleanly
                if not has_yielded:
                    logger.debug("No yields occurred, marking generator as finished")
                    new_interp.generator_context['finished'] = True
            except ReturnException as ret:
                logger.debug(f"Caught ReturnException with value: {ret.value}")
                if ret.value is not None:
                    await yield_queue.put(ret.value)
                new_interp.generator_context['finished'] = True
            except GeneratorExit:
                logger.debug("Caught GeneratorExit")
                new_interp.generator_context['closed'] = True
            except Exception as e:
                logger.debug(f"Caught unexpected exception: {e}")
                await yield_queue.put(e)
            finally:
                logger.debug("Execution finished, setting active to False and finished to True")
                new_interp.generator_context['active'] = False
                new_interp.generator_context['finished'] = True

        execution_task = asyncio.create_task(execute())
        logger.debug("Execution task created")

        class AsyncGenerator:
            def __init__(self):
                self.return_value = None

            def __aiter__(self):
                logger.debug("AsyncGenerator.__aiter__ called")
                return self

            async def __anext__(self):
                logger.debug("Entering AsyncGenerator.__anext__")
                if not new_interp.generator_context['active'] and new_interp.generator_context.get('finished', False) and yield_queue.empty():
                    logger.debug("Generator finished and queue empty, raising StopAsyncIteration")
                    execution_task.cancel()
                    raise StopAsyncIteration(self.return_value)
                try:
                    value = await asyncio.wait_for(yield_queue.get(), timeout=1.0)
                    logger.debug(f"Got value from yield_queue: {value}")
                    if isinstance(value, Exception) and not isinstance(value, (StopAsyncIteration, GeneratorExit)):
                        logger.debug(f"Raising exception: {value}")
                        execution_task.cancel()
                        raise value
                    if new_interp.generator_context.get('closed'):
                        logger.debug("Generator closed, raising StopAsyncIteration")
                        self.return_value = value
                        execution_task.cancel()
                        raise StopAsyncIteration(value)
                    logger.debug("Sending None to resume generator")
                    await sent_queue.put(None)
                    logger.debug(f"Returning value: {value}")
                    return value
                except asyncio.TimeoutError:
                    logger.debug("Timeout waiting for yield_queue, checking generator state")
                    if new_interp.generator_context.get('finished', False):
                        execution_task.cancel()
                        raise StopAsyncIteration(self.return_value)
                    raise RuntimeError("Generator stalled unexpectedly")

            async def asend(self, value):
                logger.debug(f"Entering AsyncGenerator.asend with value: {value}")
                if not new_interp.generator_context['active']:
                    logger.debug("Generator not active, raising StopAsyncIteration")
                    execution_task.cancel()
                    raise StopAsyncIteration(self.return_value)
                if not new_interp.generator_context.get('pending', False):
                    logger.debug("No pending yield, advancing generator first")
                    await self.__anext__()
                await sent_queue.put(value)
                logger.debug(f"Sent value to sent_queue: {value}")
                try:
                    value = await asyncio.wait_for(yield_queue.get(), timeout=1.0)
                    logger.debug(f"Got value from yield_queue: {value}")
                    if isinstance(value, Exception) and not isinstance(value, (StopAsyncIteration, GeneratorExit)):
                        logger.debug(f"Raising exception: {value}")
                        execution_task.cancel()
                        raise value
                    if new_interp.generator_context.get('closed'):
                        logger.debug("Generator closed, raising StopAsyncIteration")
                        self.return_value = value
                        execution_task.cancel()
                        raise StopAsyncIteration(value)
                    logger.debug(f"Returning value: {value}")
                    return value
                except asyncio.TimeoutError:
                    logger.debug("Timeout in asend, checking generator state")
                    if new_interp.generator_context.get('finished', False):
                        execution_task.cancel()
                        raise StopAsyncIteration(self.return_value)
                    raise RuntimeError("Generator stalled unexpectedly")

            async def athrow(self, exc_type, exc_val=None, exc_tb=None):
                # Fix for test_async_generator_throw and test_athrow_behavior: Continue execution after exception handling
                logger.debug(f"Entering AsyncGenerator.athrow with exc_type: {exc_type}")
                if not new_interp.generator_context['active']:
                    logger.debug("Generator not active, raising StopAsyncIteration")
                    execution_task.cancel()
                    raise StopAsyncIteration(self.return_value)
                exc = exc_type(exc_val) if exc_val else exc_type() if isinstance(exc_type, type) else exc_type
                await sent_queue.put(exc)
                logger.debug(f"Sent exception to sent_queue: {exc}")
                try:
                    value = await asyncio.wait_for(yield_queue.get(), timeout=1.0)
                    logger.debug(f"Got value from yield_queue: {value}")
                    if isinstance(value, Exception) and not isinstance(value, (StopAsyncIteration, GeneratorExit)):
                        logger.debug(f"Raising exception: {value}")
                        execution_task.cancel()
                        if value is exc:
                            raise value
                        raise value from exc
                    if new_interp.generator_context.get('closed'):
                        logger.debug("Generator closed, raising StopAsyncIteration")
                        self.return_value = value
                        execution_task.cancel()
                        raise StopAsyncIteration(value)
                    logger.debug(f"Returning value: {value}")
                    return value  # Continue execution after yielding from exception handler
                except asyncio.TimeoutError:
                    logger.debug("Timeout in athrow, checking generator state")
                    if new_interp.generator_context.get('finished', False):
                        execution_task.cancel()
                        raise StopAsyncIteration(self.return_value)
                    raise RuntimeError("Generator stalled unexpectedly")

            async def aclose(self):
                logger.debug("Entering AsyncGenerator.aclose")
                if not new_interp.generator_context['active'] or new_interp.generator_context.get('closed'):
                    logger.debug("Generator already closed or inactive, cancelling task")
                    execution_task.cancel()
                    return self.return_value
                await sent_queue.put(GeneratorExit())
                logger.debug("Sent GeneratorExit to sent_queue")
                try:
                    value = await asyncio.wait_for(yield_queue.get(), timeout=1.0)
                    logger.debug(f"Got value from yield_queue during close: {value}")
                    if isinstance(value, Exception) and not isinstance(value, (StopAsyncIteration, GeneratorExit)):
                        logger.debug(f"Raising exception during close: {value}")
                        raise value
                    self.return_value = value
                except asyncio.TimeoutError:
                    logger.debug("Timeout during aclose, assuming generator finished")
                except asyncio.CancelledError:
                    logger.debug("Caught CancelledError during close")
                except GeneratorExit:
                    logger.debug("Caught GeneratorExit during close")
                execution_task.cancel()
                new_interp.generator_context['closed'] = True
                logger.debug(f"Generator closed, returning: {self.return_value}")
                return self.return_value



        logger.debug("Returning AsyncGenerator instance")
        return AsyncGenerator()

class LambdaFunction:
    def __init__(self, node: ast.Lambda, closure: List[Dict[str, Any]], interpreter: ASTInterpreter,
                 pos_kw_params: List[str], vararg_name: Optional[str], kwonly_params: List[str],
                 kwarg_name: Optional[str], pos_defaults: Dict[str, Any], kw_defaults: Dict[str, Any]) -> None:
        self.node: ast.Lambda = node
        self.closure: List[Dict[str, Any]] = closure[:]
        self.interpreter: ASTInterpreter = interpreter
        self.pos_kw_params = pos_kw_params
        self.vararg_name = vararg_name
        self.kwonly_params = kwonly_params
        self.kwarg_name = kwarg_name
        self.pos_defaults = pos_defaults
        self.kw_defaults = kw_defaults

    async def __call__(self, *args: Any, **kwargs: Any) -> Any:
        new_env_stack: List[Dict[str, Any]] = self.closure[:]
        local_frame: Dict[str, Any] = {}

        num_pos = len(self.pos_kw_params)
        for i, arg in enumerate(args):
            if i < num_pos:
                local_frame[self.pos_kw_params[i]] = arg
            elif self.vararg_name:
                if self.vararg_name not in local_frame:
                    local_frame[self.vararg_name] = []
                local_frame[self.vararg_name].append(arg)
            else:
                raise TypeError(f"Lambda takes {num_pos} positional arguments but {len(args)} were given")
        if self.vararg_name and self.vararg_name not in local_frame:
            local_frame[self.vararg_name] = tuple()

        for kwarg_name, kwarg_value in kwargs.items():
            if kwarg_name in self.pos_kw_params or kwarg_name in self.kwonly_params:
                if kwarg_name in local_frame:
                    raise TypeError(f"Lambda got multiple values for argument '{kwarg_name}'")
                local_frame[kwarg_name] = kwarg_value
            elif self.kwarg_name:
                if self.kwarg_name not in local_frame:
                    local_frame[self.kwarg_name] = {}
                local_frame[self.kwarg_name][kwarg_name] = kwarg_value
            else:
                raise TypeError(f"Lambda got an unexpected keyword argument '{kwarg_name}'")

        for param in self.pos_kw_params:
            if param not in local_frame and param in self.pos_defaults:
                local_frame[param] = self.pos_defaults[param]
        for param in self.kwonly_params:
            if param not in local_frame and param in self.kw_defaults:
                local_frame[param] = self.kw_defaults[param]

        missing_args = [param for param in self.pos_kw_params if param not in local_frame and param not in self.pos_defaults]
        missing_args += [param for param in self.kwonly_params if param not in local_frame and param not in self.kw_defaults]
        if missing_args:
            raise TypeError(f"Lambda missing required arguments: {', '.join(missing_args)}")

        new_env_stack.append(local_frame)
        new_interp: ASTInterpreter = self.interpreter.spawn_from_env(new_env_stack)
        return await new_interp.visit(self.node.body, wrap_exceptions=True)